/*********************************************************************/
/* newsucc.c                                                         */
/* function neblo_newsucc(), manages a recursive answer to a         */
/* previous "findsucc" cmd.                                          */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


int 
neblo_newsucc
( reqst_t *reqst )
{
  
  char                 *keymod;
  char                 *keyexp;
  u_int32_t             handle_val;
  unsigned int          port;
  ULL                   a;
  neblo_addr            low_bound;
  neblo_addr            old_up_bound;
  peer_t               *f;
  int                   rc;
  unsigned int          i;


  /*  SECURITY:  accept only if we must replace a dead successor  */
  f = successor();
  if (f->status != DEAD)
    return -1;

  _do_(_malloc_(keymod,KEYBUFLEN+SLACK),
       !keymod,"neblo_newsucc(): keymod malloc");
  _do_(_malloc_(keyexp,BUFLEN+SLACK),
       !keyexp,"neblo_newsucc(): keyexp malloc");

  sscanf(reqst->buf,"%*s %u %u %016llx %50000s %100s",
         &handle_val,&port,&a,keymod,keyexp);
  low_bound = a;

  /*  Be sure no other thread is doing the same.  */
  cmd_handle_lock(&f->handle);

  /*  SECURITY II  */
  if (cmd_handle_is_done(&f->handle)) {

#ifdef DEBUG_M
    S(fprintf(stderr,
              "[%lu]\tneblo_newsucc(): findsucc cmd handle not found\n",
              reqst->th_id));
#endif

    goto abort;
  }

  /*  SECURITY III  */
  if (cmd_handle_val(&f->handle) != handle_val) {

#ifdef DEBUG_M
    S(fprintf(stderr,
              "[%lu]\tneblo_newsucc(): bad findsucc cmd handle\n",
              reqst->th_id));
#endif

    goto abort;
  }

  /*  Prepare to fill peer desc with info about new successor  */

  /*
   *  Cleaning the old successor here is not harmful for routing, as
   *  its status was DEAD so it was ignored by the routing algorithm.
   *  After cleaning, the status becomes NONE so harmless as well.
   *  Status becomes TOCALL upon successful end of this routine,
   *  otherwise it is set back to DEAD.
   */
  clean_peer_desc_nolock(f);  /*  this also cleans the cmd handle  */

  f->addr = reqst->peer_addr;
  f->port = (u_int16_t)port;

  _do_(_RSA_new_(f->pubkey),!f->pubkey,
       "neblo_newsucc(): RSA_new of f->pubkey");
  _convert_pubkey_(f->pubkey,keymod,keyexp,/* */);
  if (!(f->pubkey->n && f->pubkey->e)) {

#ifdef DEBUG_M
    S(fprintf(stderr, "\tbad pubkey from new successor\n"));
#endif

    goto abort;
  }

  /*  New successor put a challenge on our pubkey.  */
  rc = _get_challenge_(reqst->csock,reqst->buf);

  if (rc < 0) {

#ifdef DEBUG_M
    S(fprintf(stderr, "\tchallenge interrupted\n"));
#endif

    goto abort;
  }

#ifdef DEBUG_M
  S(fprintf(stderr, "\tchallenge done\n"));
#endif


  /*  Wait new successor to be ready to accept our friendship  */

  _read_(reqst->csock,reqst->buf,BUFLEN,rc,TIMEOUT,goto abort);
  if (strncmp(reqst->buf,"ok",strlen("ok"))) goto abort;

#ifdef DEBUG
  S(fprintf(stderr, "\t>> %s\n",reqst->buf));
#endif

  /*  Now establish the connection to new successor  */

  for (i=5; i; i--) {
    rc = call_peer(f,1 /* from predecessor */,0,reqst->th_id);
    if (rc >= 0)
      /* success. successor->status is set to OK */
      break;
    if (rc == -2) {
      /* successor has died meanwhile. abort... */

#ifdef DEBUG_M
      S(fprintf(stderr,"\tsuccessor dead. abort...\n"));
#endif

      goto abort;
    }
    /* new successor not responding. wait a moment, then try again */
    sleep(1);
  }
  if (!i) {
    /* successor not responding for too long.  assume it is dead */

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\tsuccessor not responding, abort...\n",
                reqst->th_id));
#endif

    goto abort;
  }

  /*
   * our new successor will
   * eventually open a connection to us (the backward connection from any
   * peer to its predecessor).
   */
  rc = insert_expd_reqst(f->addr,f->port,2 /* from successor */,f->pubkey,
                         reqst->th_id);

  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto abort);

#ifdef DEBUG
  S(fprintf(stderr, "\t<< ok\n"));
#endif

  /*
   *  The incoming connection will be managed by a separate thread.
   *  No need to wait for this.  Done.
   */

  pthread_mutex_lock(&ring->lock);
  f->status = TOCALL;
  old_up_bound = ring->up_bound;
  ring->up_bound = low_bound - 1;
  pthread_mutex_unlock(&ring->lock);

  ask_refresh_addrs(old_up_bound,ring->up_bound,NEBLO_MAXCP,reqst->th_id);

#ifdef DEBUG_M
  S(fprintf(stderr,
            "[%lu]\tneblo_newsucc(): new successor: %u.%u.%u.%u:%u\n",
            reqst->th_id,
            ((unsigned char *)&f->addr.s_addr)[0],
            ((unsigned char *)&f->addr.s_addr)[1],
            ((unsigned char *)&f->addr.s_addr)[2],
            ((unsigned char *)&f->addr.s_addr)[3],
            f->port));
#endif

  rc = 0;

exit:

  _free_(keymod); _free_(keyexp);

  if (rc >= 0)
    cmd_handle_reset(&f->handle);

  cmd_handle_unlock(&f->handle);
  f->status = DEAD;

  return rc;

abort:

  rc = -1;
  goto exit;

}
